-- Hazara Khan's implementation of the Suicide Kings loot distribution
--  system. See http://www.wowwiki.com/Suicide_Kings for a description
--  of the system, and <Seraphim> of Medivh's loot rules for the basis
--  for this implementation.
-- This is intended to be a module for HazLoot, a master looter aid, but
--  written in such a way that it's testable outside of WoW.

-- TODO: With Capitalize, consider a HSK.DisplayName table

debug = false

HSK = {}
if not HazLootList then -- not in WoW, because Ace creates HazLootList
  HSK.standalone = true
end

-- DATA STRUCTURES
local MasterList   = {} -- The master list    {[index] = member}
local MembersByName= {} -- index of {[name] = member}
local InvertedList = {} -- Inverted master list {[memberName]=index} -- replace?
local Alts         = {} -- Alts of members' mains in the master list
                        --  {[memberName]={alt1,alt2,...altN,}}
local Mains        = {} -- Mains of members' alts (inverse of Alts)
local Present      = {} -- A sub-list of the master, of present raiders' mains
local InvertedP    = {} -- Inverted present list {[main]=presentKey} (mains are lowercase)
local PresentCount = 0  -- Present is not 1,2,..n-indexed, have to count size
local Roles        = {} -- Raiders' declared roles {[name]=role}, includes alts.
local RoleSet      = {} -- Set of permissible roles {[role]=count} where count is of Mains

-- Members
--Member prototype:
local Member = {
  present  = false,
  position = 0,
  suicides = 0,
  offspecs = 0,
  name     = "",
  role     = {},
  alts     = {},
  new = function(self, o)
          o = o or {}
          self.__index = self
          o.suicides = 0
          o.offspecs = 0
          return o
        end,
}

-- this copy is a hack. Move it somewhere else!
local STATE  = {}
local STATES = {}
STATE.IDLE              = 0 -- Scanning, looting, whatever. Not bidding.
STATE.AUCTION_OPEN      = 1 -- Auction is are open
STATE.LASTCALL          = 2 -- Last call for bids
STATE.ROLL_OPEN         = 3 -- Bids are closed - no valid bids - gone to offspec rolls
STATE.BANKING           = 4 -- Item is bing banked
STATE.DUSTING           = 5 -- Item is bing dusted
STATE.CONFIRMED         = 6 -- Bid winner is being confirmed
STATE.AWARDING          = 7 -- Loot is being awarded
STATES = {"Bid open","Last call","Roll open","Banking","Dusting","Confirming bid winner","Awarding loot!",}
STATES[STATE.IDLE] = "Idle"

-- OPTIONS
local DropOnMainChange = true -- This is Seraphim's default
local DropOnRoleChange = true -- This is Seraphim's default

--- Player names have initial capitals, and all lowercase thereafter.
local function Capitalize(unit)
  if nil == unit then return nil end
  if "string" ~= type(unit) then return unit end
  unit = string.upper(string.sub(unit,1,1))..string.lower(string.sub(unit,2))
  return unit
end

local function GetMemberByName(name)
  local cName = Capitalize(name)
  return MembersByName[cName]
end

--- Replace the master list with an ordered list of raiders.
-- @param members An array of the members of an SK list, in suicide order, with their roles:
--                {"member1","role1","member2","role2",..."memberN","roleN"}
-- @param alts (Optional) A table of [member]={alt1,alt2,...altN,} entries. Keys in the alts
--                        table must exist in the members array parameter. May be nil or {}
-- @return The number of members added if the list was created correctly, nil otherwise
function HSK:SetMasterList(members, alts)
  if not "table"==type(members) then return false end
  if alts and not "table" == type (alts) then return false end
  HSK:Clear()
  for i,main in ipairs(members) do
    if 1 == i%2 then
      main = Capitalize(main)
      HSK:AddRole(members[i+1])
      HSK:AddMember(main, members[i+1])
    end
  end
  if alts then
    for member, altList in pairs(alts) do
      for i,alt in ipairs(altList) do
        if 1 == i%2 then
          alt = Capitalize(alt)
          HSK:AddRole(altList[i+1])
          HSK:AddAlt(member, alt,altList[i+1])
        end
      end
    end
  end
  return (HSK:GetMasterListSize() > 0)
end

-- -- -- STRATEGY INTERFACE -- -- -- --
function HSK:GetName()
  return "Suicide Kings (Seraphim)"
end

-- TODO: Implement this to handle DropOnMainChange, DropOnRoleChange etc
function HSK:GetStrategyOptions()
  return nil
end

--- Initialize the strategy using the HazLoot database reference
-- @returns number of raiders if succeeded, false if failed (and should be bootstrapped)
function HSK:Init(dbref)
  local mList = nil
  local aList = nil
  if dbref and "table" == type(dbref) then 
    mList = dbref.profile.MasterList
    aList = dbref.profile.AltList
  else
    print("DB ref not a table. FAIL! dbref is a "..type(dbref))
    return false
  end
  if mList then
    return HSK:SetMasterList(mList, aList)
  else
    print("mList is "..(type(mList) or "nil"))
    print("dbref is "..(type(dbref) or "nil"))
    print("dbref.profile.MasterList is "..(type(dbref.profile.MasterList) or "nil"))
  end
  print("Hunh? FAIL!")
  return false
end

function HSK:SaveState(dbref)
  local masterList = {}
  local altList = {}

  if MasterList then 
    for i,member in ipairs(MasterList) do
      main = member.name
      table.insert(masterList, main)
      table.insert(masterList, Roles[main])
      if Alts[main] then
        altList[main] = {}
        for j,alt in ipairs(Alts[main]) do
          table.insert(altList[main],alt)
          table.insert(altList[main],Roles[alt])
        end
      end
    end
  end

  if masterList then
    self:Debug("Saving master list")
    dbref.profile.MasterList = masterList 
  end
  if altList then
    self:Debug("Saving alts list")
    dbref.profile.AltList = altList 
  end

  return true
end

HSK.auction = "suicide"
HSK.roll    = "off-spec roll"
HSK.bank    = "guild banking"
HSK.dust    = "disenchanting"

--- returns true if state change is valid, usage message otherwise
function HSK:IsValidStateChange(from, to)
  if from == to then return true
  elseif (to == STATE.AUCTION_OPEN  and from == STATE.IDLE) or
     (to == STATE.AUCTION_OPEN  and from == STATE.AUCTION_OPEN) or
     (to == STATE.AUCTION_OPEN  and from == STATE.LASTCALL) or
     (to == STATE.AUCTION_OPEN  and from == STATE.ROLL_OPEN) or
     (to == STATE.LASTCALL      and from == STATE.AUCTION_OPEN) or
     (to == STATE.ROLL_OPEN     and from == STATE.LASTCALL) or
     (to == STATE.CONFIRMED     and from == STATE.LASTCALL) or
     (to == STATE.DUSTING       and from == STATE.LASTCALL) or
     (to == STATE.BANKING       and from == STATE.LASTCALL) or
     (to == STATE.CONFIRMED     and from == STATE.ROLL_OPEN) or
     (to == STATE.DUSTING       and from == STATE.ROLL_OPEN) or
     (to == STATE.BANKING       and from == STATE.ROLL_OPEN) or
     (to == STATE.DUSTING       and from == STATE.BANKING) or
     (to == STATE.BANKING       and from == STATE.DUSTING) or
     (to == STATE.DUSTING       and from == STATE.IDLE) or
     (to == STATE.BANKING       and from == STATE.IDLE) or
     (to == STATE.CONFIRMED     and from == STATE.BANKING) or
     (to == STATE.CONFIRMED     and from == STATE.DUSTING) or
     (to == STATE.AUCTION_OPEN  and from == STATE.CONFIRMED) or
     (to == STATE.AWARDING      and from == STATE.CONFIRMED) or
     (to == STATE.LASTCALL      and from == STATE.CONFIRMED) or
     (to == STATE.BANKING       and from == STATE.CONFIRMED) or
     (to == STATE.DUSTING       and from == STATE.CONFIRMED) or
     (to == STATE.ROLL_OPEN     and from == STATE.CONFIRMED)
    then
      -- self:Debug("Valid going from: "..STATES[from].." to "..STATES[to])
      return true
  else
      self:Debug("NOT valid going from: "..STATES[from].." to "..STATES[to])
  end
  return HSK:GetNextStates(from)
end

function HSK:GetAnnouncement(state)
  local message = ""
  return message
end

function HSK:GetNextStates(state)
  local usage = "dunno"
  if state == STATE.IDLE then
    usage = "/hl |cff1eff00n|rext to open bids on the next item"
  elseif state == STATE.AUCTION_OPEN then
    usage = "/hl |cff1eff00l|rast|cff1eff00c|rall to issue a last call"
  elseif state == STATE.LASTCALL then
    usage =  "/hl |cff1eff00n|rext to go re-open the auction\n"
    usage = usage.."/hl |cff1eff00o|rff|cff1eff00s|rpec to go to off-spec rolls\n"
    usage = usage.."/hl |cff1eff00w|rinner <name1>, <name2>,... to set the suicide list winner\n"
    usage = usage.."/hl |cff1eff00b|rank to send the item to the banker\n"
    usage = usage.."/hl |cff1eff00d|rust to send the item to the disenchanter"
  elseif state == STATE.BANKING then
    usage = " /hl |cffff1e00a|rward to bank "
  elseif state == STATE.DUSTING then
    usage = " /hl |cffff1e00a|rward to dust "
  elseif state == STATE.CONFIRMED then
    self:Print(" /hl |cff1eff00a|rward to distribute confirmed loot")
  end
  return usage
end

function HSK:GetHelp(state)
  local help = "I'm confused. Try reloading your UI with /rl"
  if state == STATE.IDLE then
    help = "click here to announce, or \'suicide\' to open bids"
  elseif state == STATE.AUCTION_OPEN then
    help = "click \'last call\' to close suicide bidding"
  elseif state == STATE.LASTCALL then
    help = "\'suicide\' to re-open, or go to \'off-spec\', \'bank\' or \'dust\'"
  elseif state == STATE.BANKING then
    help = "click \'confirm\' to send to the banker"
  elseif state == STATE.DUSTING then
    help = "click \'confirm\' to send to the duster"
  elseif state == STATE.CONFIRMED then
    help = "click \'confirm\' to send to the winner"
  end
  return help
end

function HSK:GetConfirmation(state)
  local confirmation = "Confirm"
  if state == STATE.LASTCALL then
    confirmation = "Suicide"
  elseif state == STATE.ROLL_OPEN then
    confirmation = "Off-spec Roll"
  elseif state == STATE.BANKING then
    confirmation = "Banker"
  elseif state == STATE.DUSTING then
    confirmation = "Disenchant"
  end
  return confirmation
end

--- Return the a candidate winner from an array of members of the Present list.
function HSK:GetCandidate(array)
  if not array then return nil end
  if not Present then return nil end
  if 0 == PresentCount then return nil end
  local winner = #MasterList + 2
  local j = nil
  for i,cName in ipairs(array) do
    array[i] = Capitalize(cName)
    if HSK:IsPresent(cName) then
      j = HSK:GetMasterPosition(cName)
      winner = (j < winner) and j or winner
    end
  end
  if winner > #MasterList then print("didn't find winner :(") end
  return MasterList[winner].name
end

--- Does a member fit a role? Some strategies restrict bids and rolls by role
-- @param member A raid member
-- @param role A role
-- @returns true if both member and role are valid and member is of role
function HSK:IsMemberInRole(member, role)
  if not member or not role then return false end
  return role == HSK:GetRole(member)
end

--- Sets the permissible roles.
-- All members must have a raid role, and each role must belong to the set of known
-- (or permissible) roles.
-- @param roles A string array of raid role names.
-- @return The number of roles added.
function HSK:SetRoles(roles)
  if not roles then return nil end
  local rolecount = 0 -- # operator FAILs on non-array tables. Crappy fucking language.
  for i,role in ipairs(roles) do
    RoleSet[role] = 0
    rolecount = rolecount + 1
  end
  return rolecount
end

--- Adds a new role
-- @param role The role to add
-- @returns true if role was added, or already exists, nil if role is invalid
function HSK:AddRole(role)
  if not role or "string" ~= type(role) then return nil end
  if RoleSet[role] then return true end
  RoleSet[role] = 0
  return nil
end

--- How many toons (mains and alts) have this role
function HSK:GetRoleCount(role)
  if not role then return 0 end
  if not RoleSet[role] then return 0 end
  return RoleSet[role]
end

--- Add a new member to the Suicide Kings list at an (optional) position
-- Attempts to append a new member to the bottom of the list, or at the
--  position provided. If position would leave a gap at the bottom of the
--  list, then the member is added at the bottom, leaving no gaps.
-- @param member   (Required) The main character name of a raider in the SK List
-- @param role     (Required) The main's declared raiding role.
-- @param position (Optional) The position in the list of the new member. All existing
--                 members at, or below, this position will be shuffled down one slot.
--                 Position must be > 0 If position is omitted, the new member starts
--                 at the bottom of the list.
-- @return The new position of the added member, nil if addition failed, the member or
--          role are undefined, or the role is not known.
function HSK:AddMember(name, role, position)
  if not name or not role then return nil end
  name = Capitalize(name)
  if not RoleSet[role] then
    self:Debug("[HSK] no such role: '"..(role or "nil").."'")
  return nil end
  if InvertedList[name] then
    return InvertedList[name]
  end
  local listsize = (#MasterList)
  member = Member:new()
  member.name = name
  MembersByName[name] = member
  if nil == position or listsize < position then -- append
    table.insert(MasterList, member)
    position = (#MasterList)
    member.position = position
    InvertedList[member.name] = position
  elseif 0 < position and position <= listsize then
    for i = position,listsize do
      InvertedList[MasterList[i].name] = i+1
    end
    table.insert(MasterList, position, member)
    member.position = position
    InvertedList[member.name] = position
  else
    return nil
  end
  Roles[member.name] = role
  member.role = role
  RoleSet[role] = RoleSet[role] + 1
  return InvertedList[member.name]
end

--- Is the toon a main or alt in this SK List?
-- @returns true if so, false if not, nil if bad input
function HSK:IsMember(name)
  if nil == name then return nil end
  if InvertedList[name] or Mains[name] then
    return true
  end
  return nil
end

--- Returns the raid role of a member (or alt)
-- @param member (Required) The character's name. May be main or alt.
-- @returns the declared raid role, nil if the member is not known.
function HSK:GetRole(member)
  if nil == member then return nil end
  member = Capitalize(member)
  return Roles[member] or Roles[HSK:GetMain(member)]
end

--- Clears the master, present and alt lists, the role lists, and role counts
-- @returns true on success
function HSK:Clear()
  MasterList   = {}
  InvertedList = {}
  Alts         = {}
  Mains        = {}
  Present      = {}
  InvertedP    = {}
  PresentCount = 0 
  Roles        = {}
  RoleSet      = {}
  return true
end

--- Remove a member from the Suicide Kings list
-- Permanently removes a member from the SK List - for example when they leave
-- the guild, stop raiding, are an utter douchebag, whatever. Adjusts the list
-- so that the list has no gaps.
-- @param member The name of a member to remove, or their position in the list.
-- @param confirm true|false If true, remove the main, and all alts corresponding
--                to member. Otherwise, return the name of the main that would
--                be nuked.
-- @return true if the member was successfully removed, the name of the main that
--         would be removed if confirm is not true, nil if not found
function HSK:RemoveMember(member, confirm)
  if not member then return false end
  local position
  local membername
  if "string" == type(member) then
    position = HSK:GetMasterPosition(member)
  elseif "number" == type(member) then
    position = member
  end
  if position >= 0 and position < (#MasterList) then
    membername = HSK:GetMemberAt(position) -- may be an alt
    if not confirm then return membername end -- must have confirmation
    if Alts[membername] then
      for i,alt in ipairs(Alts[membername]) do
        HSK:RemoveAlt(alt)
      end
    end
    Alts[membername] = nil
    RoleSet[Roles[membername]] = RoleSet[Roles[membername]] - 1
    Roles[membername] = nil
    InvertedList[membername] = nil
    local listsize = (#MasterList)
    for i = position+1,listsize do
      MasterList[i-1] = MasterList[i]
      InvertedList[MasterList[i].name] = i-1
    end
    table.remove(MasterList, listsize)
    return true
  end
  return false
end

--- Add an alt to the SK List, associated with its corresponding main.
--  @param main The name of the correpsonding main
--  @param alt  The name of the alt in question
--  @param role (Optional) The role of the alt in question, defaults to main's role
--  @return The position of the main in the SK List, nil if the main was
--          not found, or the alt is already on another main, or an invalid
--          role was specified
function HSK:AddAlt(main, alt, role)
  if not main or not alt then return nil end
  main = Capitalize(main)
  alt  = Capitalize(alt)
  if not role then role = HSK:GetRole(main) end
  if not RoleSet[role] then return nil end
  local mainPos = HSK:GetMasterPosition(main)
  local altPos  = HSK:GetMasterPosition(alt)
  if altPos and mainPos == altPos then return altPos -- Already exists
  elseif altPos then -- Alt exists on another main
    return nil
  end
  if mainPos then 
    if not Alts[main] then Alts[main] = {alt}
    else
      table.insert(Alts[main],alt)
    end
    Mains[alt] = main
    Roles[alt] = role
    RoleSet[role] = RoleSet[role] + 1
  end
  return mainPos
end

--- Remove an alt from the records, leaving the main unaffected
--  @param alt The name of the alt in question
--  @return true if the alt was removed, false if the alt was a
--          main, nil if the alt was not found
function HSK:RemoveAlt(alt)
  if not alt then return nil end
  alt = Capitalize(alt)
  if InvertedList[alt] then return false end
  local main = Mains[alt]
  if not main then return nil end
  for i,anAlt in ipairs(Alts[main]) do
    if alt == anAlt then
      RoleSet[Roles[alt]] = RoleSet[Roles[alt]] - 1
      table.remove(Alts[main],i)
      Mains[alt] = nil
      return true
    end
  end
  return nil
end

--- Promote an alt to main, and move the corresponding main to the alt list.
-- @param alt The name of an alt to promote
-- @return The position of the new main in the SK List. In Seraphim, changing main
--         means a drop to the bottom of the list. This is configurable. Returns nil
--         if no alt is specified. No-op if given the name of a main.
-- @see SetDropOnMainChange
function HSK:PromoteAlt(alt)
  if nil == alt then return nil end
  alt = Capitalize(alt)
  local oldMain = HSK:GetMain(alt)
  if alt == oldMain then return InvertedList[oldMain] end -- Already a main
  local newMain = alt
  -- Set inverted list for new main
  InvertedList[newMain] = InvertedList[oldMain]
  -- Switch them in the Mains table
  Mains[newMain] = nil
  Mains[oldMain] = newMain
  -- Fill newMain's alt table with oldMain's alt table
  Alts[newMain] = {oldMain}
  for i,alt in ipairs (Alts[oldMain]) do
    if newMain ~= alt then
      table.insert(Alts[newMain],alt)
    end
  end
  -- Remove reference to oldMain's alt table
  Alts[oldMain] = nil
  -- For each member, set its main to newMain, noting which was newMain
  local newMainOldIndex = 0
  for i,anAlt in ipairs(Alts[newMain]) do
    if newMain == anAlt
      then newMainOldIndex = i
    else
      Mains[anAlt] = newMain
    end
  end

  -- If we drop on main or role change, drop
  if DropOnMainChange then
    return HSK:Drop(newMain)
  elseif DropOnRoleChange and Roles[oldMain] ~= Roles[newMain] then
    return HSK:Drop(newMain)
  else
    return InvertedList[newMain]
  end
  return nil
end

--- Changes the raid role of a main or alt
-- Depending on how the SK List is configured, changing a main's role may drop the
--  main to the bottom of the SK List, by default. Nothing happens if the new role
--  is the same as the old role.
-- @param member The toon changing role
-- @param role   The new role
-- @returns The new role, or nil if member was invalid or not found, or the role was unknown
-- @see SetDropOnRoleChange
function HSK:ChangeRole(member,role)
  if not member then return nil end
  local main = HSK:GetMain(member)
  local oldRole = Roles[member]
  if role == oldRole then return oldRole -- no change
  else
    RoleSet[oldRole] = RoleSet[oldRole] - 1
    Roles[member]    = role
    RoleSet[role]    = RoleSet[role] + 1
  end
  if main == member and DropOnRoleChange then
    HSK:Drop(member)
  end
end

--- Returns an array of a member's alts.
-- @param main The main whose alts you want
-- @returns A cloned array of the member's alts, empty if there are none,
--          nil if main is invalid, not found or an alt
function HSK:GetAlts(main)
  if not main then return nil end
  if not InvertedList[main] then return nil end
  local alts = {}
  if Alts[main] then 
    for i,alt in ipairs(Alts[main]) do
      table.insert(alts, alt)
    end
  end
  return alts
end

--- Returns the main of a member (alt or main)
function HSK:GetMain(member)
  if not member then return nil end
  member = Capitalize(member)
  if InvertedList[member] then return member end
  return Mains[member]
end

--- Return the number of known toons (Main + Alts) of a member
-- @param member The name of an alt or main.
-- @returns Number of known toons (1 + #Alts) for member, nil if
--          member is nil, 0 if member is not known
function HSK:GetToonCount(member)
  if not member then return nil end -- bad input
  local mainPos = HSK:GetMasterPosition(member)
  if not mainPos then return 0 end  -- not found
  local alts = Alts[MasterList[mainPos].name]
  if not alts then return 1 end
  return 1 + #alts
end

--- Set an existing member to "present", and thus eligible for suicides
-- @param member The name of a member to mark as present, or an alt
-- @returns true if the member is present, the name of an alt's main if member is
--          an alt or false if the member is not present
function HSK:MarkAsPresent(name)
  if not name then return nil end
  if "Unknown" == name then return nil end
  local member = GetMemberByName(name)
  local main = HSK:GetMain(member.name)
  member.present = true
  if nil == main then return nil end
  if InvertedP[member.name] then return true
  elseif InvertedP[main] then return main end
  InvertedP[main] = InvertedList[main]
  Present[InvertedP[main]] = main
  PresentCount = PresentCount + 1
  return main
end

--- Is the member present in the raid?
-- @param member The name of a main or alt
-- @returns true if the member is present, the name of an alt's main if member is
--          an alt or false if the member is not present
function HSK:IsPresent(member)
  if not member then return nil end
  member = Capitalize(member)
  local main = HSK:GetMain(member)
  if not main then return nil end
  if main ~= member then return main end
  return (nil ~= InvertedP[main] and true or nil)
end

--- Return a member's current position in the master SK List
--  @param position Integer position in the list.
--  @return The name of the member in the position provided, or nil if position
--          is larger than the size of the list.
function HSK:GetMemberAt(position)
  return MasterList[position].name
end

--- Return the current position in the master list of a raid member
--  @param member A main in the master list, or one of their alts.
--  @return The position in the master list of a raid member, or the position in
--          the master list of a main of one of that member's alts. Returns nil
--          if member is not in the master list, nor an alt.
function HSK:GetMasterPosition(member)
  if nil == member then return nil end
  member = Capitalize(member)
  local main = Mains[member]
  return InvertedList[member] or InvertedList[main]
end

function HSK:HandleWinner(name, cost)
  print("handling winner: "..(name or "nil"))
  if not name then return nil end
  member = GetMemberByName(name)
  if not Present or 0 == PresentCount then print("nobody present - weird") return nil end
  if cost > 0 then
    member.suicides = member.suicides + 1
    return HSK:Drop(HSK:GetMain(member.name), true)
  else
    member.offspecs = member.offspecs + 1
  end
end

--- Drops an active member to the bottom of the Master SK List
--  @param member The main or alt of a main that's being dropped
--  @param list   The list in which to drop member (Master if nil)
--  @return their new position in the list, nil if they were not found.
function HSK:Drop(name, usePresent)
  print("dropping "..name)
  member = GetMemberByName(name)
  usePresent = usePresent or false
  if usePresent then print("in the 'present' list") end

  local i = member.position
  local nextPos = i+1
  local safety = member.position
  local old = nil
  local new = nil
  while i <= #MasterList and safety <= #MasterList do
    old = MasterList[i]
    new = MasterList[nextPos] or member
    if not usePresent or new.present then
      MasterList[i] = new
      lastPos = new.position
      new.position = i
      i = nextPos
    end
    nextPos = nextPos + 1
    safety = safety + 1
  end

  -- rebuild the inverted list
  InvertedList = {}
  for i,member in ipairs(MasterList) do
    InvertedList[member.name] = i
  end
  -- rebuild the present list, if there was one
  if (PresentCount > 0)  then
    local tmpPresent = {}
    for i,present in pairs(Present) do
      table.insert(tmpPresent, present)
    end
    Present = {}
    InvertedP = {}
    PresentCount = 0
    for i,present in pairs(tmpPresent) do
      HSK:MarkAsPresent(present)
    end
  end
  return InvertedList[name]
end

--- How many people are in the master list?
--  @return The size of the master list, 0 if empty.
function HSK:GetMasterListSize()
  return (#MasterList)
end

--- How many people are in the "present" list?
--  @return The size of the present list, 0 if empty.
function HSK:GetPresentCount()
  return (PresentCount)
end

--- Returns (cloned) arrays of the master and alts list, with their roles.
--  The lists returned can be persisted externally for subsequent restoration with
--   HSK:SetMasterList()
--  @return Cloned copies of the master SK table and the corresponding Alts with roles
--  @see HSK:SetMasterList
function HSK:GetMasterList()
  local list = {}
  if MasterList then 
    for i,member in pairs(MasterList) do
      local main = member.name
      table.insert(list,{["name"]    = main,
                         ["main"]    = main,
                         ["roles"]   = HSK:GetRole(main),
                         ["present"] = (nil ~= HSK:IsPresent(main)),
                         ["won"]     = (member.suicides>0), -- To be implemented
                         ["master"]  = i,
                        })
    end
  end
  return list
end

function HSK:GetPresentList()
  local list = {}
  if Present then 
    for i,main in pairs(Present) do
      local member = GetMemberByName(main)
      table.insert(list,{["name"]    = main,
                         ["main"]    = HSK:GetMain(main),
                         ["roles"]   = HSK:GetRole(main),
                         ["present"] = true,
                         ["won"]     = (member.suicides>0), -- To be implemented
                         ["master"]  = i,
                        })
    end
  end
  table.sort(list, function (a,b) return (a.master < b.master) end)
  return list
end


--- Export the master list to some format
-- By default, returns a string of CSV, for each main: "Position","Main","Role","Alt1","Alt1Role","Alt2",..
-- TODO: write a corresponding import
-- @params format Currently ignored (defaults to CSV)
function HSK:ExportMasterList(format)
  local export = ""
  for i,member in ipairs(MasterList) do
    local main = member.name
    export = export.."\""..i.."\",\""..main.."\",\""..HSK:GetRole(main).."\""
    if Alts[main] then
      for j,alt in ipairs(Alts[main]) do
        export = export..",\""..alt.."\",\""..HSK:GetRole(alt).."\""
      end
    end
    export = export.."\n"
  end
  return export
end

--- Export the present list to some format
-- By default, returns a string of CSV, for each main: "Position","Main","Role"
-- @params format Currently ignored (defaults to CSV)
function HSK:ExportPresentList(format)
  local export = {}
  local output = ""
  if Present then
    for i,main in pairs(Present) do
      table.insert(export,"\""..i.."\",\""..main.."\",\""..HSK:GetRole(main).."\"".."\n")
    end
    table.sort(export)
    for i,line in ipairs(export) do
      output = output..line
    end
  else
    output = "no members present"
  end
  return output
end

--- Bootstrap the SKList from Bootstrap.lua
--  To start using HazLoot's SK implementation, you can either edit Bootstrap.lua or use the
--  default, empty list and the (as yet unimplemented) UI to add people manually
-- @params tableSet Name of tables in Bootstrap.lua to use to bootstrap
-- @returns true if succesfully bootstrapped, false otherwise
function HSK:Bootstrap(tableSet)
  if not tableSet then return false end
  local masterList, altList = Bootstrap:GetBootTables(tableSet)
  HSK:SetMasterList(masterList, altList)
  return false
end

--- Should the module drop a member to the bottom of the list when changing mains
function HSK:GetDropOnMainChange()
  return DropOnMainChange
end
--- Set whether the module should drop a member to the bottom of the list when changing mains
function HSK:SetDropOnMainChange(boolean)
  if nil == boolean then return nil end
  DropOnMainChange = boolean
end

--- Should the module drop a member to the bottom of the list when changing roles
function HSK:GetDropOnRoleChange()
  return DropOnRoleChange
end
--- Set whether the module should drop a member to the bottom of the list when changing roles
function HSK:SetDropOnRoleChange(boolean)
  if nil == boolean then return nil end
  DropOnRoleChange = boolean
end

function HSK:Print(message)
  if HazLoot then
    HazLoot:Print(message)
  else
    print("[HSK] "..message)
  end
end

function HSK:Debug(message)
  if HazLoot then
    HazLoot:Debug(message)
  else
    print("[HSK][DEBUG] "..message)
  end
end

--- Public Test functions ----

function HSK:TestCapitalize()
  -- Capitalize
  assert("Unit" == Capitalize("unit"), "Capitalization failed")
  assert("Unit" == Capitalize("UNIT"), "Capitalization failed")
  assert("Unit" == Capitalize("UnIt"), "Capitalization failed")
  return true
end
